@@ -34,9 +34,7 @@
 
 
 int
-ebl_abi_cfi (ebl, abi_info)
-     Ebl *ebl;
-     Dwarf_CIE *abi_info;
+ebl_abi_cfi (Ebl *ebl, Dwarf_CIE *abi_info)
 {
   return ebl == NULL ? -1 : ebl->abi_cfi (ebl, abi_info);
 }
